"""Console scripts for threedigrid."""


import click
from IPython.terminal.embed import InteractiveShellEmbed

from threedigrid.admin.gridadmin import GridH5Admin
from threedigrid.management import help_texts

file_type_choice_map = {"shape": "to_shape", "gpkg": "to_gpkg"}


@click.command()
@click.option(
    "--grid-file",
    prompt="Path to the admin file",
    help="Path to the admin file",
    type=click.Path(exists=True),
)
@click.option(
    "--ipy/--no-ipy", default=False, help="Start an interactive ipython session"
)
def kick_start(grid_file, ipy):
    grid = GridH5Admin(grid_file)

    txt = help_texts.model_overview(grid_file)
    click.secho("""\nOverview of model specifics:\n""", bold=True)
    click.secho(txt)

    if ipy:
        info_txt = (
            "\n\nA GridH5Admin instance for model {} has been "
            "created (variable named grid). Call dir(grid) to "
            "explore.\n".format(grid.model_name)
        )
        click.secho(info_txt, bg="green", fg="black", bold=True)
        ipshell = InteractiveShellEmbed(exit_msg="Ciao...\n")
        ipshell()


@click.command()
@click.option(
    "--grid-file",
    prompt="Path to the admin file",
    help="Path to the admin file",
    type=click.Path(exists=True),
)
@click.option("--file-type", type=click.Choice(["shape", "gpkg"]))
@click.option(
    "--output-file",
    prompt="Path to the output file",
    help="Path to the output file",
    type=click.Path(writable=True, resolve_path=True),
)
@click.option("--model", type=click.Choice(["nodes", "lines", "breaches", "levees"]))
@click.option(
    "--subset",
    default="",
    help="Filter by a subset like 1D_all " "(applies only for models nodes and lines",
)
def export_to(grid_file, file_type, output_file, model, subset):
    success_msg = "Successfully created {} ".format(output_file)
    grid = GridH5Admin(grid_file)
    m = getattr(grid, model)
    export_func = file_type_choice_map[file_type]
    if subset:
        getattr(m.subset(subset), export_func)(output_file)
        click.secho(success_msg, fg="green", bold=True)
        return
    getattr(m, export_func)(output_file)
    click.secho(success_msg, fg="green", bold=True)


if __name__ == "__main__":
    kick_start()



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.

from collections import OrderedDict

LONLAT_DIGITS = 7  # 7 decimals is about 11mm accuracy

TYPE_V2_PIPE = "v2_pipe"
TYPE_V2_CHANNEL = "v2_channel"
TYPE_V2_CULVERT = "v2_culvert"
TYPE_V2_ORIFICE = "v2_orifice"
TYPE_V2_WEIR = "v2_weir"
TYPE_V2_MANHOLE = "v2_manhole"
TYPE_V2_PUMPSTATION = "v2_pumpstation"
TYPE_V2_LEVEE = "v2_levee"
TYPE_V2_1D_LATERAL = "v2_1d_lateral"
TYPE_V2_WINDSHIELD = "v2_windshielding"

TYPE_V2_1D_BOUNDARY_CONDITIONS = "v2_1d_boundary_conditions"
TYPE_V2_2D_BOUNDARY_CONDITIONS = "v2_2d_boundary_conditions"
TYPE_V2_CONNECTION_NODES = "v2_connection_nodes"

TYPE_V2_BREACH = "v2_breach"
TYPE_V2_CROSS_SECTION_DEF = "v2_cross_section_definition"
TYPE_V2_CROSS_SECTION_LOCATION = "v2_cross_section_location"
TYPE_V2_ADDED_CALCULATION_POINT = "v2_added_calculation_point"


TYPE_CODE_MAP = {
    TYPE_V2_PIPE: 1,
    TYPE_V2_CHANNEL: 2,
    TYPE_V2_CULVERT: 3,
    TYPE_V2_ORIFICE: 4,
    TYPE_V2_WEIR: 5,
    TYPE_V2_MANHOLE: 6,
    TYPE_V2_PUMPSTATION: 7,
    TYPE_V2_LEVEE: 8,
    TYPE_V2_1D_LATERAL: 9,
    TYPE_V2_1D_BOUNDARY_CONDITIONS: 10,
    TYPE_V2_2D_BOUNDARY_CONDITIONS: 11,
    TYPE_V2_CONNECTION_NODES: 12,
    TYPE_V2_BREACH: 13,
    TYPE_V2_CROSS_SECTION_DEF: 14,
    TYPE_V2_CROSS_SECTION_LOCATION: 15,
    TYPE_V2_ADDED_CALCULATION_POINT: 16,
    TYPE_V2_WINDSHIELD: 17,
}

# ------------------ HDF5 NAMES ------------------ #
DSET_ID_MAPPING = "id_map"
GROUP_MAPPINGS = "mappings"

DSET_GEOMS = "geoms"
GROUP_LEVEE = "levees"

###############################################################################
# HDF5 ATTRIBUTE NAMES
EXTENT_1D_KEY = "extent_1d"
EXTENT_2D_KEY = "extent_2d"
THREEDICORE_VERSION_KEY = "threedicore_version"
HAS_1D_KEY = "has_1d"
HAS_2D_KEY = "has_2d"
HAS_PUMPSTATIONS_KEY = "has_pumpstations"
HAS_BREACHES_KEY = "has_breaches"


###############################################################################
# MANHOLE_TYPES is used by:
#
#   - v2_manhole

MANHOLE_TYPES = {
    0: "manhole",
    1: "outlet",
    2: "pumpstation",
}


CALCULATION_TYPES = {
    # -1 tot 99 ::= Node | 'verbinding-zonder-subnodes(="vertices")'
    -1: "boundary node",
    0: "embedded",
    1: "stand-alone",
    2: "connected",
    3: "broad crested",  # only orifices + weirs, corresponds
    4: "short crested",  # only orifices + weirs
    5: "double connected",
    # 100 tot oneindig ::= 'verbinding-met-subnodes(="vertices")'
    # a.k.a. pinpoint
    100: "embedded",
    101: "stand-alone",
    102: "connected",
    105: "double connected",
}

# Hardcoded zoom levels for pipe/nod/nod_1d/line/pump
# for 1-5 zoom level ("zoom_category") in sqlite file.


ZOOM_LEVELS = {
    "pipe": list(range(16, 21)),
    "nod": list(range(16, 21)),
    "nod_1d": list(range(16, 21)),
    "line": list(range(16, 21)),
    "pump": list(range(16, 21)),
}


###############################################################################
# SEWERAGE_TYPES is used by:
#
#   - v2_pipe

SEWERAGE_TYPES = {
    0: "combined",
    1: "stormwater",
    2: "wastewater",
    3: "transport",
    4: "overflow",
    5: "sinker",
    6: "storage",
    7: "storage settling tank",
    8: None,
}


###############################################################################
# CREST_TYPES is used by:
#
#   - v2_weir
#   - v2_orifice
#
# NB! This mapping is related to CALCULATION_TYPES mapping (key=3 | key=4)

CREST_TYPES = {
    3: "broad crested",  # only orifices + weirs
    4: "short crested",  # only orifices + weirs
}

# accounts for 2D only
LINE_BASE_FIELDS = OrderedDict(
    [
        ("link_id", "int"),
        ("kcu", "int"),
        ("kcu_descr", "str"),
        ("node_a", "int"),
        ("node_b", "int"),
    ]
)

# extends LINE_BASE_FIELDS in case of 1D
LINE_1D_FIELDS = OrderedDict(
    [
        ("cont_type", "str"),
        ("cont_pk", "int"),
        ("seq_id", "int"),
    ]
)

LINE_BASE_FIELDS_ALL = OrderedDict(LINE_BASE_FIELDS, **LINE_1D_FIELDS)

# maps the fields names of grid line objects
# to their external representation
LINE_FIELD_NAME_MAP = OrderedDict(
    [
        ("link_id", "id"),
        ("kcu", "kcu"),
        ("kcu_descr", "kcu_descr"),
        ("node_a", "node_a"),
        ("node_b", "node_b"),
        ("seq_id", "lik"),
        ("cont_type", "content_type"),
        ("cont_pk", "content_pk"),
    ]
)

# accounts for 2D only
NODE_BASE_FIELDS = OrderedDict(
    [
        ("nod_id", "int"),
    ]
)

NODE_1D_FIELDS = OrderedDict(
    [
        ("con_nod", "str"),
        ("con_nod_pk", "int"),
    ]
)

NODE_FIELD_NAME_MAP = OrderedDict(
    [
        ("nod_id", "id"),
        ("con_nod", "seq_id"),
        ("con_nod_pk", "content_pk"),
    ]
)

FID_FIELDS = {"id", "nod_id", "link_id"}


def to_string(value):
    # special handling of bytes (else you get the string "b'something'")
    if isinstance(value, bytes):
        return value.decode()
    else:
        return str(value)


def to_float(value):
    if value is None:
        return None
    try:
        value = float(value)
    except ValueError:
        return None
    # value == value checks if it is not NaN
    return value if (value != -9999.0) and (value == value) else None


def to_int(value):
    if value is None:
        return None
    try:
        value = int(value)
    except ValueError:
        return None
    return value


TYPE_FUNC_MAP = {
    "int": int,
    int: to_int,
    float: to_float,
    str: to_string,
    bool: lambda x: None if x in (-9999, "-9999") else bool(x),
    "float": to_float,
    "str": to_string,
}


###############################################################################
# shapefile names (served by tilestache)
SUBSET_1D_ALL = "1D_all"
SUBSET_2D_GROUNDWATER = "2D_groundwater"
SUBSET_2D_OPEN_WATER = "2D_open_water"
SUBSET_2D_VERTICAL_INFILTRATION = "2D_vertical_infiltration"

# grids
_BASE_NAME = "grid_"
GROUNDWATER = _BASE_NAME + SUBSET_2D_GROUNDWATER
OPEN_WATER = _BASE_NAME + SUBSET_2D_OPEN_WATER

# nodes
_NODES_BASE = "nodes_"
NODES = _NODES_BASE + SUBSET_1D_ALL

# lines
_LINES_BASE = "lines_"
LINES = _LINES_BASE + SUBSET_1D_ALL
GROUNDWATER_LINES = _LINES_BASE + SUBSET_2D_GROUNDWATER
OPEN_WATER_LINES = _LINES_BASE + SUBSET_2D_OPEN_WATER
VERTICAL_INFILTRATION_LINES = _LINES_BASE + SUBSET_2D_VERTICAL_INFILTRATION

# levees
LEVEES = "levees"

# Flow (1d2d) lines
FLOWLINES = "flowlines"

###############################################################################
# for extent calculations

# key: name of the node subset
# value: hdf5 attribute
SUBSET_NAME_H5_ATTR_MAP = {
    SUBSET_1D_ALL.upper(): EXTENT_1D_KEY,
    SUBSET_2D_OPEN_WATER.upper(): EXTENT_2D_KEY,
}

NO_DATA_VALUE = -9999.0

# the default slice for result timeseries
DEFAULT_CHUNK_TIMESERIES = slice(0, 10)



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.


import logging
from abc import ABCMeta
from collections import OrderedDict
from itertools import chain, tee

import numpy as np

from threedigrid.orm.base.exceptions import OperationNotSupportedError
from threedigrid.orm.base.fields import (
    ArrayField,
    IndexArrayField,
    TimeSeriesArrayField,
)
from threedigrid.orm.base.filters import get_filter, SliceFilter
from threedigrid.orm.base.options import Options

logger = logging.getLogger(__name__)


def extend_instance(obj, cls):
    """Apply mixins to a class instance after creation"""
    base_cls = obj.__class__
    base_cls_name = obj.__class__.__name__
    obj.__class__ = type(base_cls_name, (base_cls, cls), {})


def pairwise(iterable):
    # from https://docs.python.org/2/library/
    # itertools.html#recipes
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


class Model(metaclass=ABCMeta):
    id = IndexArrayField()

    _field_names = []

    _exporters = []

    _filter_map = None

    _datasource = None

    def __init__(
        self,
        datasource=None,
        slice_filters=[],
        epsg_code=None,
        only_fields=[],
        reproject_to_epsg=None,
        has_1d=None,
        mixin=None,
        timeseries_chunk_size=None,
        **kwargs
    ):
        """
        Initialize a Model with a datasource, filters
        and a epsg_code.

        The datasource is a wrapped h5py Group from a H5py file or
        an instance that implements the same interface.
        """

        self._datasource = datasource
        self.slice_filters = slice_filters
        self.only_fields = only_fields
        self.reproject_to_epsg = reproject_to_epsg

        self.class_kwargs = {
            "slice_filters": slice_filters,
            "only_fields": only_fields,
            "reproject_to_epsg": reproject_to_epsg,
            "has_1d": has_1d,
            "mixin": mixin,
            "timeseries_chunk_size": timeseries_chunk_size,
        }

        # Extend the class with the mixin, if set
        if mixin:
            extend_instance(self, mixin)
            # pass kwargs to mixin
            super().__init__(**kwargs)

        # Cache the boolean filter mask for this instance
        # after it has been computed once
        self._boolean_mask_filter = None
        self._mixin = mixin
        self._epsg_code = epsg_code

        # Cache the field names
        _field_names = [
            x
            for x in dir(self.__class__)
            if isinstance(
                getattr(self.__class__, x), (ArrayField, TimeSeriesArrayField)
            )
        ]
        self._field_names = set(self._field_names).union(set(_field_names))

        self.has_1d = has_1d

    @property
    def epsg_code(self):
        if not self._epsg_code:
            self._epsg_code = self._datasource.getattr("epsg_code")
        return self._epsg_code

    @property
    def count(self):
        """count of all elements (including trash element)"""
        if self.class_kwargs.get("slice_filters"):
            return self.get_filtered_field_value("id").size
        return self.get_field_value("id").size

    @property
    def model_name(self):
        try:
            model_name = "-".join(
                (
                    self._datasource.getattr("model_name"),
                    self._datasource.getattr("model_slug"),
                    str(self._datasource.getattr("revision_nr")),
                    self._datasource.getattr("revision_hash"),
                )
            )
        except (AttributeError, KeyError, TypeError):
            model_name = "unknown"
            pass
        return model_name

    def _get_field(self, field_name):
        """
        Returns: the ArrayField with field_name on this instance.
        """
        return super().__getattribute__(field_name)

    def get_field_value(self, field_name, **kwargs):
        """
        Returns: the value for the ArrayField with name field_name
        on the instance.

        With this in place we can later override the ArrayField
        get_value method for field specific conversion, if needed.
        """
        update_dict = {
            "model_name": self.__class__.__name__,
        }

        kwargs.update(update_dict)
        return self._meta.get_field(field_name).get_value(
            self._datasource, field_name, **kwargs
        )

    def _get_subset_idx(self, field_name):
        """
        get an array of indexes for the given subset

        :param field_name: field name
        """
        new_inst = self.__init_class(
            self.__class__, **{"mixin": self.class_kwargs.get("mixin")}
        )
        subset_dict = new_inst.Meta.subset_fields.get(field_name)
        if not subset_dict:
            return
        _subset_name = list(subset_dict.keys())
        if not _subset_name:
            return
        return new_inst.subset(_subset_name[0]).id

    def get_filtered_field_value(
        self, field_name, ts_filter=None, lookup_index=None, subset_index=None
    ):
        # Redirect via datasource
        return self._datasource.get_filtered_field_value(
            self, field_name, ts_filter, lookup_index, subset_index
        )

    def __getattribute__(self, attr_name):
        """
        Override the __getattribute__ methode to return
        the computed values of ArrayField's instead of
        the ArrayField
        instance.
        """
        # Note: don't use getattr(self, XX) in this function,
        # it will lead to a max recursion error
        attr = super().__getattribute__(attr_name)
        # Override getting ArrayField attributes
        if isinstance(attr, ArrayField):
            # Return meta directly, we don't want it to
            # to be filtered
            if attr_name == "meta":
                return attr.get_value(self._datasource, attr_name)

            # Use the get function to retrieve the computed/filtered
            # value for the ArrayField with name: 'name'
            return super().__getattribute__("get_filtered_field_value")(attr_name)

        # Default behaviour, return the attribute from superclass
        return attr

    def __init_class(self, klass, **kwargs):
        """
        Returns: a new instance of 'klass' with new filters and
                 same datasource.
        """

        return klass(datasource=self._datasource, **kwargs)

    def __get_filters(self, **kwargs):
        """
        Parse kwargs for filters.

        For example:
            content_type__in=[1,2,3]

            splitted_key[0] = "content_type"
            splitted_key[1] = "in"
            -> Try to find filter with "in" as key

        """
        new_slice_filters = list(self.slice_filters)  # make copy
        filter_as = kwargs.pop("filter___as", False)

        for key, value in kwargs.items():
            # python2/3 combat
            if isinstance(value, str):
                value = str.encode(value)
            splitted_key = key.split("__")
            if splitted_key[0] not in self._field_names:
                raise ValueError(
                    "Field '{}' unknown. Choices are {}".format(
                        splitted_key[0], self._field_names
                    )
                )

            new_slice_filters.append(
                get_filter(
                    splitted_key,
                    self._meta.get_field(splitted_key[0]),
                    value,
                    filter_map=self._filter_map,
                    filter_as=filter_as,
                )
            )

        return new_slice_filters

    def _filter_as(self, klass, **kwargs):
        """
        Same as self.filter, but now return "klass" instance instead of
        self.__class__ instance
        """
        # Inject filter___as, set this on the filters
        # so they can identified as filters that
        # produce submodels
        kwargs["filter___as"] = True
        slice_filters = self.__get_filters(**kwargs)
        new_class_kwargs = dict(self.class_kwargs)
        new_class_kwargs.update({"slice_filters": slice_filters})

        return self.__init_class(klass, **new_class_kwargs)

    # def _set_subsets(self):
    #     if not hasattr(self, 'Meta'):
    #         return
    #     if not hasattr(self.Meta, 'subset_fields'):
    #         return
    #
    #     _subsets = set(self.Meta.subset_fields.keys())
    #     for sset in _subsets:
    #         idx = self._get_subset_idx(sset)
    #         setattr(self, sset, idx)

    def filter(self, **kwargs):
        """
        Django style filtering, allows multiple filter or chaining filters
        for example:

            line_instance.filter(content_type='v2_weir', content_pk=1)
            line_instance.filter(content_type='v2_weir').filter(content_pk=1)

            Both produce the same result object.


        Returns: a new instance of the object with the extra filters added.
        """
        slice_filters = self.__get_filters(**kwargs)
        new_class_kwargs = dict(self.class_kwargs)
        new_class_kwargs.update({"slice_filters": slice_filters})

        return self.__init_class(self.__class__, **new_class_kwargs)

    def slice(self, start, stop=None, step=None, override_filter_error=False):
        """
        slice by name or slice. See instance.predefined_slices for an overview
        of predefined slices

        :param s: either the name of the slice (see instance.predefined_slices
             for an overview or an slice instance, e.g. slice(<x>, <y>)
        :return: nodes instance with containing the sliced dataset.Example
            usage::
            usage:
            open_water_2d = nodes.slice('2D_open_water')
            open_water_2d.data

                OrderedDict([('content_pk', array([1, 2, 3, ...

        """
        if self.slice_filters and not override_filter_error:
            raise OperationNotSupportedError(
                "You cannot use slices on a filtered dataset"
            )

        if isinstance(start, slice):
            slice_filter = start
        else:
            slice_filter = slice(start, stop, step)

        new_class_kwargs = dict(self.class_kwargs)
        new_class_kwargs.update(
            {"slice_filters": [SliceFilter(slice_filter)] + self.slice_filters}
        )

        return self.__init_class(self.__class__, **new_class_kwargs)

    @property
    def known_subset(self):
        if not hasattr(self, "SUBSETS"):
            return "has no subsets defined"
        return list(chain(*[list(v.keys()) for v in self.SUBSETS.values()]))

    def subset(self, name):
        """
        get a subset of the data. See instance.subsets for an overview
        of predefined subsets

        :param name: name of the subset (see instance.predefined_slices
             for an overview or an slice instance, e.g. slice(<x>, <y>)
        :return: nodes instance with containing the sliced dataset.Example
            usage::
            open_water_2d = nodes.slice('2D_open_water')
            open_water_2d.data

                OrderedDict([('content_pk', array([1, 2, 3, ...

        """
        if not hasattr(self, "SUBSETS"):
            raise TypeError("SUBSETS not defined for this type of model")

        if isinstance(name, str):
            field_filter = [
                key for key in self.SUBSETS if name.upper() in self.SUBSETS[key]
            ]

            if not field_filter:
                msg = "Invalid subset name."
                msg += " Valid options are: {} ".format(
                    [x for y in self.SUBSETS.values() for x in y],
                )
                logger.exception(msg)
                raise KeyError(msg)

            if len(field_filter) > 1:
                raise KeyError("Multiple options found for subset: %s " % name.upper())

            field_filter = field_filter[0]
        else:
            raise TypeError(
                "Type %s not supported"
                % type(
                    name,
                )
            )

        return self.filter(**{field_filter: self.SUBSETS[field_filter][name.upper()]})

    def only(self, *args):
        """
        Return only the fields requested

        Usage: line_instance.only('content_pk', 'line')
        """
        if not args:
            raise Exception("Please provide at least one field name")

        for x in args:
            if x not in self._field_names:
                raise Exception("Unknown field name: {}".format(x))

        new_class_kwargs = dict(self.class_kwargs)
        new_class_kwargs.update({"only_fields": self.only_fields + list(args)})

        return self.__init_class(self.__class__, **new_class_kwargs)

    def _get_values(self, selection):
        """
        when no filters are specified, we are still operating on the hdf5
        groups so get the values explicitly

        :param selection: ordered dict of selected fields
            with group datasets
        :return: new ordered dict with values from group datasets
        """

        _tmp = OrderedDict()
        for k, v in selection.items():
            try:
                _tmp[k] = v[:]
            except TypeError:
                pass
        return _tmp

    def to_dict(self):
        """
        Returns: the filtered values as dictionairy
        """
        return self._datasource.execute_query(self)

    def to_structured_array(self):
        """
        :return: the filtered values as a
        structured (named) array
        """
        selection = self.to_dict()

        # Convert the dictionary to structured array
        dtypes = []
        for key, value in selection.items():
            dtypes.append((key, value.dtype.name, value.shape))

        return np.array([tuple(selection[x[0]] for x in dtypes)], dtype=dtypes)[0]

    def to_list(self):
        """
        :return: list of dicts with key's and values
        """

        # Filter results and transform the result to
        # and np.ndarray
        selection = self.to_dict()
        if len(list(selection.values())) > 1:
            array = np.array(list(selection.values()))
        else:
            array = list(selection.values())[0]

        def optional_zip(array_to_zip):
            """
            Optional zip the results of
            array_to_zip if there are more than one
            dimensions.
            """
            # Transform data to list
            array_as_list = array_to_zip.tolist()
            if len(array_to_zip.shape) > 1:
                # Zip the list if there are more than
                # one dimensions.
                array_as_list = zip(*array_as_list)
            return array_as_list

        # Zip the lists of array_lists (does same as a transpose)
        data = zip(*[optional_zip(x) for x in array])

        # Create a list of dictionairies of the data by
        # zipping the selection.keys() (= field_names) for all items
        # in data
        return [dict(zip(list(selection.keys()), x)) for x in data]

    @property
    def boolean_mask_filter(self):
        selection = OrderedDict()

        # Compute the boolean mask filter
        # that can be applied to all array's
        # based on the specified filters

        if self._boolean_mask_filter is None:
            # If no filters are defined
            if not self.slice_filters:
                # Select everything
                return slice(None)

            # Only load the fields that need to be filtered
            filter_field_names = [x.get_field_name() for x in self.slice_filters]

            for name in [x for x in filter_field_names if x]:
                selection[name] = self.get_field_value(name)

            boolean_mask_filter = None
            # Compute the the filter
            for filter_instance in self.slice_filters:
                filter_bool_mask = filter_instance.get_boolean_mask_filter(
                    selection, self
                )

                if boolean_mask_filter is None:
                    boolean_mask_filter = filter_bool_mask
                else:
                    # apply the filter_bool_mask on the boolean_mask_filter
                    # where the boolean_mask_filter is still "True".
                    boolean_mask_filter[
                        np.equal(boolean_mask_filter, True)
                    ] &= filter_bool_mask  # noqa

            self._boolean_mask_filter = boolean_mask_filter

        return self._boolean_mask_filter

    def to_array(self):
        """
        Returns: the filtered values as a numpy array
        """
        selection = self.to_dict()
        if len(list(selection.values())) > 1:
            return np.array(list(selection.values()))
        return list(selection.values())[0]

    @property
    def data(self):
        """
        Returns: the (filtered) data a dictionairy
        """
        return self.to_dict()

    @property
    def _meta(self):
        """
        meta class that add meta data to a ResultMixin instance. The meta data
        includes entries for the ``_meta_fields`` collection (if found).

            >>> from threedigrid.admin.gridresultadmin import GridH5ResultAdmin
            >>> f = "/code/tests/test_files/results_3di.nc"
            >>> ff = "/code/tests/test_files/gridadmin.h5"
            >>> gr = GridH5ResultAdmin(ff, f)
            >>> gr.nodes._meta.s1
            >>> s1(units=u'm', long_name=u'waterlevel', standard_name=u'water_surface_height_above_reference_datum') # noqa
        """
        return Options(self)

    def __repr__(self):
        """human readable representation of the instance"""
        return "<orm {} instance of {}>".format(self.__contenttype__(), self.model_name)

    def __contenttype__(self):
        """conent type, e.g. lines, nodes, cells, ..."""
        return self._datasource.group_name



import csv

from threedigrid.admin.gridresultadmin import (
    _GridH5NestedStructureControl,
    GridH5StructureControl,
)
from threedigrid.admin.structure_controls.models import StructureControlTypes


def structure_control_actions_to_csv(
    structure_control: GridH5StructureControl, out_path: str
):
    """Set place table, timed, and memory controls after each other in one file."""
    with open(out_path, "w", newline="") as csvfile:
        csv_writer = csv.writer(csvfile, delimiter=",")
        csv_writer.writerow(
            [
                "control_type",
                "id",
                "source_table",
                "source_table_id",
                "time",
                "action_type",
                "action_value_1",
                "action_value_2",
                "is_active",
            ]
        )
        for control_type in StructureControlTypes.__members__.values():
            control_type_data: _GridH5NestedStructureControl = getattr(
                structure_control, control_type.name
            )
            for (
                id,
                grid_id,
                time,
                action_type,
                action_value_1,
                action_value_2,
                is_active,
            ) in zip(
                control_type_data.id,
                control_type_data.grid_id,
                control_type_data.time,
                control_type_data.action_type,
                control_type_data.action_value_1,
                control_type_data.action_value_2,
                control_type_data.is_active,
            ):
                source_table, source_table_id = structure_control.get_source_table(
                    action_type, grid_id
                )
                csv_writer.writerow(
                    [
                        control_type.value,
                        id,
                        source_table,
                        source_table_id,
                        time,
                        action_type,
                        action_value_1,
                        action_value_2,
                        is_active,
                    ]
                )



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.

import logging
import os
from collections import OrderedDict

try:
    from osgeo import ogr
except ImportError:
    ogr = None

from threedigrid.admin import exporter_constants as const
from threedigrid.admin.constants import (
    NODE_1D_FIELDS,
    NODE_BASE_FIELDS,
    NODE_FIELD_NAME_MAP,
)
from threedigrid.geo_utils import get_spatial_reference
from threedigrid.orm.base.exporters import BaseOgrExporter

logger = logging.getLogger(__name__)


class NodesOgrExporter(BaseOgrExporter):
    """
    Exports to ogr formats. You need to set the driver explicitly
    before calling save()
    """

    def __init__(self, nodes):
        """
        :param lines: lines.models.Lines instance
        """
        self._nodes = nodes
        self.supported_drivers = {
            const.GEO_PACKAGE_DRIVER_NAME,
            const.SHP_DRIVER_NAME,
            const.GEOJSON_DRIVER_NAME,
        }

    def save(self, file_name, node_data, target_epsg_code, **kwargs):
        """
        save to file format specified by the driver, e.g. shapefile

        :param file_name: name of the outputfile
        :param node_data: dict of node data
        """
        assert self.driver is not None
        geomtype = 0
        sr = get_spatial_reference(target_epsg_code)
        self.del_datasource(file_name)
        data_source = self.driver.CreateDataSource(file_name)
        layer = data_source.CreateLayer(str(os.path.basename(file_name)), sr, geomtype)
        fields = NODE_BASE_FIELDS
        if self._nodes.has_1d:
            fields.update(NODE_1D_FIELDS)

        for field_name, field_type in fields.items():
            layer.CreateField(
                ogr.FieldDefn(str(field_name), const.OGR_FIELD_TYPE_MAP[field_type])
            )
        _definition = layer.GetLayerDefn()

        for i in range(node_data["id"].size):
            if node_data["id"][i] == 0:
                continue  # skip the dummy element
            point = ogr.Geometry(ogr.wkbPoint)
            point.AddPoint(
                node_data["coordinates"][0][i], node_data["coordinates"][1][i]
            )
            feature = ogr.Feature(_definition)
            feature.SetGeometry(point)
            for field_name, field_type in fields.items():
                fname = NODE_FIELD_NAME_MAP[field_name]
                try:
                    raw_value = node_data[fname][i]
                except IndexError:
                    raw_value = None
                self.set_field(feature, field_name, field_type, raw_value)
            layer.CreateFeature(feature)


class CellsOgrExporter(BaseOgrExporter):
    """
    Exports to ogr formats. You need to set the driver explicitly
    before calling save()
    """

    def __init__(self, cells):
        """
        :param lines: lines.models.Lines instance
        """
        self._cells = cells
        self.supported_drivers = {
            const.GEO_PACKAGE_DRIVER_NAME,
            const.SHP_DRIVER_NAME,
        }
        self.driver = None

    def save(self, file_name, cells_data, target_epsg_code, **kwargs):
        """
        save to file format specified by the driver, e.g. shapefile

        :param file_name: name of the outputfile
        :param cells_data: dict of cell data
        """
        assert self.driver is not None

        geomtype = 0
        sr = get_spatial_reference(target_epsg_code)

        self.del_datasource(file_name)
        data_source = self.driver.CreateDataSource(file_name)
        layer = data_source.CreateLayer(str(os.path.basename(file_name)), sr, geomtype)
        fields = OrderedDict(
            [
                ("nod_id", "int"),
                ("bottom_lev", "float"),
            ]
        )
        for field_name, field_type in fields.items():
            layer.CreateField(
                ogr.FieldDefn(str(field_name), const.OGR_FIELD_TYPE_MAP[field_type])
            )

        _definition = layer.GetLayerDefn()
        for i in range(cells_data["id"].size):
            if cells_data["id"][i] == 0:
                continue  # skip the dummy element
            feature = ogr.Feature(_definition)
            # Create ring
            ring = ogr.Geometry(ogr.wkbLinearRing)
            ring.AddPoint(
                cells_data["cell_coords"][0][i], cells_data["cell_coords"][1][i]
            )

            ring.AddPoint(
                cells_data["cell_coords"][2][i], cells_data["cell_coords"][1][i]
            )

            ring.AddPoint(
                cells_data["cell_coords"][2][i], cells_data["cell_coords"][3][i]
            )

            ring.AddPoint(
                cells_data["cell_coords"][0][i], cells_data["cell_coords"][3][i]
            )

            ring.AddPoint(
                cells_data["cell_coords"][0][i], cells_data["cell_coords"][1][i]
            )

            # Create polygon from ring
            poly = ogr.Geometry(ogr.wkbPolygon)
            poly.AddGeometry(ring)
            feature.SetGeometry(poly)
            self.set_field(feature, "nod_id", "int", cells_data["id"][i])
            self.set_field(
                feature, "bottom_lev", "float", cells_data["z_coordinate"][i]
            )
            layer.CreateFeature(feature)



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.

KCU__IN_SUBSETS = {
    "1D": [0, 1, 2, 3, 4, 5],
    "1D2D": [51, 52, 53, 54, 55, 56, 57, 58],
    "2D_OPEN_WATER": [100, 101, 200, 300, 400, 500],
    "LONG_CRESTED_STRUCTURES": [3],
    "SHORT_CRESTED_STRUCTURES": [4],
    "2D_OPEN_WATER_OBSTACLES": [101],
    "2D_VERTICAL_INFILTRATION": [150],
    "2D_GROUNDWATER": [-150, 600, 700, 800, 900],
    "ACTIVE_BREACH": [56],
    "POTENTIAL_BREACH": [55],
}

KCU__IN_SUBSETS["1D_ALL"] = KCU__IN_SUBSETS["1D"] + KCU__IN_SUBSETS["1D2D"]
KCU__IN_SUBSETS["GROUNDWATER_ALL"] = (
    KCU__IN_SUBSETS["2D_GROUNDWATER"] + KCU__IN_SUBSETS["2D_VERTICAL_INFILTRATION"]
)
KCU__IN_SUBSETS["2D_ALL"] = (
    KCU__IN_SUBSETS["GROUNDWATER_ALL"]
    + KCU__IN_SUBSETS["2D_OPEN_WATER_OBSTACLES"]
    + KCU__IN_SUBSETS["2D_OPEN_WATER"]
)



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.

import logging

import numpy as np

from threedigrid.admin.prepare_utils import db_objects_to_numpy_array_dict
from threedigrid.admin.utils import PKMapper

logger = logging.getLogger(__name__)


class PreparePumps:
    @staticmethod
    def prepare_datasource(h5py_file, threedi_datasource, pump_group):
        node_group = h5py_file["nodes"]
        datasource = pump_group

        # Step 1: Load data from threedi_datasource
        pumpstations_field_names = [
            "pk",
            "display_name",
            "type",
            "start_level",
            "lower_stop_level",
            "capacity",
            "connection_node_start_pk",
            "connection_node_end_pk",
            "zoom_category",
        ]

        pumpstations_numpy_array_dict = db_objects_to_numpy_array_dict(
            threedi_datasource.v2_pumpstations, pumpstations_field_names
        )

        ids = np.array([False])
        for field_name in pumpstations_field_names:
            data = pumpstations_numpy_array_dict[field_name]
            dataset_name = field_name

            if dataset_name == "pk":
                dataset_name = "content_pk"
                ids = np.arange(1, data.size + 1)

            # insert trash element
            data = np.insert(data.copy(), 0, 0, axis=len(data.shape) - 2)

            datasource.set(dataset_name, data)

        if np.any(ids):
            data = np.insert(ids.copy(), 0, 0, axis=len(data.shape) - 2)
            datasource.set("id", data)

        # Step 2: Postprocessing data

        # Connection nodes are nodes with content_pk != 0
        connection_nodes_mask = node_group["content_pk"][:] != 0
        connection_nodes_content_pks = node_group["content_pk"][:][
            connection_nodes_mask
        ]
        connection_nodes_content_id = node_group["id"][:][connection_nodes_mask]
        connection_nodes_content_coordinates = node_group["coordinates"][:][
            :, connection_nodes_mask
        ]

        connection_nodes_initial_waterlevel = node_group["initial_waterlevel"][:][
            connection_nodes_mask
        ]

        cn1_mapper = PKMapper(
            connection_nodes_content_pks, datasource["connection_node_start_pk"]
        )
        cn2_mapper = PKMapper(
            connection_nodes_content_pks, datasource["connection_node_end_pk"]
        )

        if "node1_id" not in list(datasource.keys()):
            node1_id = cn1_mapper.apply_on(connection_nodes_content_id, -9999)
            datasource.set("node1_id", node1_id)

        if "node2_id" not in list(datasource.keys()):
            node2_id = cn2_mapper.apply_on(connection_nodes_content_id, -9999)
            datasource.set("node2_id", node2_id)

        if "bottom_level" not in list(datasource.keys()):
            bottom_level = cn1_mapper.apply_on(connection_nodes_initial_waterlevel, 0)
            datasource.set("bottom_level", bottom_level)

        if "node_coordinates" not in list(datasource.keys()):
            node1_coordinates = cn1_mapper.apply_on(
                connection_nodes_content_coordinates, -9999
            )
            node2_coordinates = cn2_mapper.apply_on(
                connection_nodes_content_coordinates, -9999
            )

            datasource.set(
                "node_coordinates", np.vstack((node1_coordinates, node2_coordinates))
            )

        if "coordinates" not in list(datasource.keys()):
            # Set the coordinates based on the centroid of node_coordinates
            # if both set and else the one that is set.
            node_coordinates = datasource["node_coordinates"][:]
            node1_coords = node_coordinates[0:2, :]
            node2_coords = node_coordinates[2:4, :]
            node1_id = datasource["node1_id"][:]
            node2_id = datasource["node2_id"][:]
            node1_not_set = node1_id == -9999
            node2_not_set = node2_id == -9999

            # Replace empty values with coords from other node
            node2_coords[:, node2_not_set] = node1_coords[:, node2_not_set]
            node1_coords[:, node1_not_set] = node2_coords[:, node1_not_set]

            # Calculate centroids
            datasource.set("coordinates", (node2_coords + node1_coords) / 2.0)



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
"""
"""


from threedigrid.orm.base.options import ModelMeta
from threedigrid.orm.base.timeseries_mixin import AggregateResultMixin, ResultMixin


class PumpsResultsMixin(ResultMixin):
    class Meta(metaclass=ModelMeta):
        field_attrs = ["units", "long_name", "standard_name"]

        composite_fields = {"q_pump": ["Mesh1D_q_pump"], "_mesh_id": ["Mesh1DPump_id"]}

    def __init__(self, **kwargs):
        """Instantiate a breach with netcdf results.

        Variables stored in the netcdf and related to breaches are dynamically
        added as attributes as TimeSeriesArrayField.

        :param kwargs:
        """
        super().__init__(**kwargs)


class PumpsAggregateResultsMixin(AggregateResultMixin):
    class Meta(metaclass=ModelMeta):
        field_attrs = ["units", "long_name"]

        model_attr = "timestamp"

        # ModelMeta will combine base_composition and composition_vars
        # to composite_fields attribute
        base_composition = {"q_pump": ["Mesh1D_q_pump"], "_mesh_id": ["Mesh1DPump_id"]}
        composition_vars = {
            "q_pump": ["min", "max", "avg", "cum", "cum_positive", "cum_negative"],
        }

    def __init__(self, **kwargs):
        """Instantiate a line with netcdf results.

        Variables stored in the netcdf and related to lines are dynamically
        added as attributes as TimeSeriesArrayField.

        :param netcdf_keys: list of netcdf variables
        :param kwargs:
        """
        super().__init__(**kwargs)



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.


import logging
import re
from collections import defaultdict
from typing import List, Optional, Union

import h5py
import numpy as np

from threedigrid.admin.breaches.models import Breaches
from threedigrid.admin.breaches.timeseries_mixin import (
    BreachesAggregateResultsMixin,
    BreachesResultsMixin,
)
from threedigrid.admin.constants import DEFAULT_CHUNK_TIMESERIES
from threedigrid.admin.gridadmin import GridH5Admin
from threedigrid.admin.h5py_datasource import H5pyResultGroup
from threedigrid.admin.h5py_swmr import H5SwmrFile
from threedigrid.admin.lines.models import Lines
from threedigrid.admin.lines.timeseries_mixin import (
    LinesAggregateResultsMixin,
    LinesResultsMixin,
)
from threedigrid.admin.nodes.models import Nodes
from threedigrid.admin.nodes.timeseries_mixin import (
    get_substance_result_mixin,
    NodesAggregateResultsMixin,
    NodesResultsMixin,
)
from threedigrid.admin.pumps.models import Pumps
from threedigrid.admin.pumps.timeseries_mixin import (
    PumpsAggregateResultsMixin,
    PumpsResultsMixin,
)
from threedigrid.admin.structure_controls.models import (
    StructureControl,
    StructureControlTypes,
)
from threedigrid.orm.models import Model

logger = logging.getLogger(__name__)

try:
    import asyncio_rpc  # noqa

    asyncio_rpc_support = True
except ImportError:
    asyncio_rpc_support = False


class GridH5ResultAdmin(GridH5Admin):
    """
    Admin interface for threedicore result queries.
    """

    def __init__(self, h5_file_path, netcdf_file_path, file_modus="r", swmr=False):
        """

        :param h5_file_path: path to the hdf5 gridadmin file
        :param netcdf_file_path: path to the netcdf result file (usually
            called subgrid_map.nc)
        :param file_modus: modus in which to open the files
        """
        self._field_model_dict = defaultdict(list)
        self._netcdf_file_path = netcdf_file_path
        super().__init__(h5_file_path, file_modus)

        self.result_datasource_class = H5pyResultGroup

        if h5_file_path.startswith("rpc://"):
            if not asyncio_rpc_support:
                raise Exception("Please reinstall this package with threedigrid[rpc]")

            from threedigrid.admin.rpc_datasource import H5RPCResultGroup, RPCFile

            self.netcdf_file = RPCFile(h5_file_path, file_modus)
            self.result_datasource_class = H5RPCResultGroup
        else:
            if swmr:
                self.netcdf_file = H5SwmrFile(netcdf_file_path, file_modus)
            else:
                self.netcdf_file = h5py.File(netcdf_file_path, file_modus)
            self.version_check()
        self.set_timeseries_chunk_size(DEFAULT_CHUNK_TIMESERIES.stop)

    def set_timeseries_chunk_size(self, new_chunk_size):
        """
        overwrite the default chunk size for timeseries queries.
        :param new_chunk_size <int>: new chunk size for
            timeseries queries
        :raises ValueError when the given value is less than 1
        """
        _chunk_size = int(new_chunk_size)
        if _chunk_size < 1:
            raise ValueError("Chunk size must be greater than 0")
        self._timeseries_chunk_size = slice(0, _chunk_size)
        logger.info("New chunk for timeseries size has been set to %d", new_chunk_size)
        self._grid_kwargs.update({"timeseries_chunk_size": self._timeseries_chunk_size})

    @property
    def timeseries_chunk_size(self):
        return self._timeseries_chunk_size.stop

    @property
    def time_units(self):
        return self.netcdf_file["time"].attrs.get("units")

    @property
    def lines(self):
        return Lines(
            self.result_datasource_class(self.h5py_file, "lines", self.netcdf_file),
            **dict(self._grid_kwargs, **{"mixin": LinesResultsMixin}),
        )

    @property
    def nodes(self):
        return Nodes(
            self.result_datasource_class(self.h5py_file, "nodes", self.netcdf_file),
            **dict(self._grid_kwargs, **{"mixin": NodesResultsMixin}),
        )

    @property
    def breaches(self):
        if not self.has_breaches:
            logger.info("Threedimodel has no breaches")
            return
        return Breaches(
            self.result_datasource_class(self.h5py_file, "breaches", self.netcdf_file),
            **dict(self._grid_kwargs, **{"mixin": BreachesResultsMixin}),
        )

    @property
    def pumps(self):
        if not self.has_pumpstations:
            logger.info("Threedimodel has no pumps")
            return
        return Pumps(
            self.result_datasource_class(self.h5py_file, "pumps", self.netcdf_file),
            **dict(self._grid_kwargs, **{"mixin": PumpsResultsMixin}),
        )

    def version_check(self):
        """
        compare versions of grid admin and grid results file.
        Issues a warning if they differ
        """
        return  # TODO implement a different check

        if self.threedicore_result_version != self.threedicore_version:
            logger.warning(
                "[!] threedicore version differ! \n"
                "Version result file has been created with: %s\n"
                "Version gridadmin file has been created with: %s",
                self.threedicore_result_version,
                self.threedicore_version,
            )

    @property
    def _field_model_map(self):
        """
        :return: a dict of {<field name>: [model name, ...]}
        """
        if self._field_model_dict:
            return self._field_model_dict

        model_names = set()
        for attr_name in dir(self):
            # skip private attrs
            if any([attr_name.startswith("__"), attr_name.startswith("_")]):
                continue
            try:
                attr = getattr(self, attr_name)
            except AttributeError:
                logger.warning(
                    "Attribute: '{}' does not " "exist in h5py_file.".format(attr_name)
                )
                continue
            if not issubclass(type(attr), Model):
                continue
            model_names.add(attr_name)

        for model_name in model_names:
            for x in getattr(self, model_name)._field_names:
                self._field_model_dict[x].append(model_name)
        return self._field_model_dict

    def get_model_instance_by_field_name(self, field_name):
        """
        :param field_name: name of a models field
        :return: instance of the model the field belongs to
        :raises IndexError if the field name is not unique across models
        """
        model_name = self._field_model_map.get(field_name)
        if not model_name or len(model_name) != 1:
            raise IndexError(
                "Ambiguous result. Field name {} yields {} model(s)".format(
                    field_name, len(model_name) if model_name else 0
                )
            )
        return getattr(self, model_name[0])

    @property
    def threedicore_result_version(self):
        """
        :return: version of the grid result file (if found), an empty
        string otherwise
        """
        try:
            return self.netcdf_file.attrs.get("threedicore_version")
        except AttributeError:
            logger.error(
                "Attribute threedicore_version could not be found in result file"
            )  # noqa
        return ""

    def close(self):
        super().close()
        self.netcdf_file.close()


class GridH5AggregateResultAdmin(GridH5ResultAdmin):
    """
    Admin interface for threedicore result queries.
    """

    def __init__(self, h5_file_path, netcdf_file_path, file_modus="r"):
        """

        :param h5_file_path: path to the hdf5 gridadmin file
        :param netcdf_file_path: path to the netcdf result file (usually
            called subgrid_map.nc)
        :param file_modus: modus in which to open the files
        """
        super().__init__(h5_file_path, netcdf_file_path, file_modus)

    @property
    def lines(self):
        model_name = "lines"
        return Lines(
            H5pyResultGroup(self.h5py_file, model_name, self.netcdf_file),
            **dict(self._grid_kwargs, **{"mixin": LinesAggregateResultsMixin}),
        )

    @property
    def nodes(self):
        return Nodes(
            H5pyResultGroup(self.h5py_file, "nodes", self.netcdf_file),
            **dict(self._grid_kwargs, **{"mixin": NodesAggregateResultsMixin}),
        )

    @property
    def breaches(self):
        if not self.has_breaches:
            logger.info("Threedimodel has no breaches")
            return
        model_name = "breaches"
        return Breaches(
            H5pyResultGroup(self.h5py_file, model_name, self.netcdf_file),
            **dict(self._grid_kwargs, **{"mixin": BreachesAggregateResultsMixin}),
        )

    @property
    def pumps(self):
        if not self.has_pumpstations:
            logger.info("Threedimodel has no pumps")
            return
        model_name = "pumps"
        return Pumps(
            H5pyResultGroup(self.h5py_file, model_name, self.netcdf_file),
            **dict(self._grid_kwargs, **{"mixin": PumpsAggregateResultsMixin}),
        )

    @property
    def time_units(self):
        logger.info("Time units are not defined globally for aggregated results")
        return None

    def close(self) -> None:
        super().close()
        self.netcdf_file.close()


class GridH5StructureControl(GridH5Admin):
    """Interface for structure control netcdf

    This interface is different from the GridH5ResultAdmin and
    GridH5AggregateResultAdmin, as it resembles less the Nodes and Lines data compared
    to those interfaces. This interface should be viewed as a CSV file interface
    which can be used to extract the different control structures by control type and
    output data. It supplies interfaces to Nodes, Lines, Weirs, etc but does not
    provide the same timeseries filter capabilities as the GridH5ResultAdmin and
    GridH5AggregateResultAdmin.

        >>> gst = GridH5StructureControl(gridadmin_path, struct_control_actions_3di.nc)
        >>> gst.table_control
        >>> gst.timed_control
        >>> gst.memory_control
        >>> gst.table_control.id
        >>> struct_control = gst.table_control.group_by_id(ga.table_control.id[0])
        >>> struct_control.time
        >>> struct_control.action_value_1
    """

    def __init__(
        self,
        h5_file_path: str,
        netcdf_file_path: str,
        file_modus: str = "r",
        swmr: bool = False,
    ):
        """
        :param h5_file_path: path to the hdf5 gridadmin file
        :param netcdf_file_path: path to the netcdf structure control result file
            (usually structure_control_actions_3di.nc)
        :param file_modus: modus in which to open the files
        """
        self._netcdf_file_path: str = netcdf_file_path
        super().__init__(h5_file_path, file_modus)

        if swmr:
            self.netcdf_file = H5SwmrFile(netcdf_file_path, file_modus)
        else:
            self.netcdf_file = h5py.File(netcdf_file_path, file_modus)

    @property
    def table_control(self) -> "_GridH5NestedStructureControl":
        return _GridH5NestedStructureControl(self, StructureControlTypes.table_control)

    @property
    def memory_control(self) -> "_GridH5NestedStructureControl":
        return _GridH5NestedStructureControl(self, StructureControlTypes.memory_control)

    @property
    def timed_control(self) -> "_GridH5NestedStructureControl":
        return _GridH5NestedStructureControl(self, StructureControlTypes.timed_control)

    def get_source_table(self, action_type, grid_id):
        """Get source_table and source_table_id based on action_type and grid_id"""
        if action_type == "set_pump_capacity":
            source_table = "v2_pumpstation"
            source_table_id = grid_id
        else:
            source_table = self.lines.content_type[grid_id].decode("utf-8")
            source_table_id = self.lines.content_pk[grid_id]

        return source_table, source_table_id


class _GridH5NestedStructureControl:
    def __init__(
        self,
        structure_control: GridH5StructureControl,
        control_type: StructureControlTypes,
    ):
        """
        :param structure_control: GridH5StructureControl(GridH5ResultAdmin)
        :param control_type: str [table_control, memory_control, timed_control]
        :param h5_file_path: path to the hdf5 gridadmin file
        :param netcdf_file_path: path to the netcdf structure control result file
            (usually structure_control_actions_3di.nc)
        :param file_modus: modus in which to open the files
        """
        if control_type not in StructureControlTypes.__members__.values():
            raise ValueError(f"Unknown control type: {control_type}")

        self.struct_control = structure_control
        self.control_type: str = control_type.value

    @property
    def action_type(self) -> np.ndarray:
        "binary character arrays from Netcdf Fortran to numpy object arrays"
        return np.array(
            [
                action_type.tobytes().decode("utf-8").strip()
                for action_type in self.struct_control.netcdf_file[
                    f"{self.control_type}_action_type"
                ][:]
            ],
            dtype=object,
        )

    @property
    def action_value_1(self) -> np.ndarray:
        return self.struct_control.netcdf_file[f"{self.control_type}_action_value_1"][:]

    @property
    def action_value_2(self) -> np.ndarray:
        return self.struct_control.netcdf_file[f"{self.control_type}_action_value_2"][:]

    @property
    def grid_id(self) -> np.ndarray:
        return self.struct_control.netcdf_file[f"{self.control_type}_grid_id"][:]

    @property
    def id(self) -> np.ndarray:
        "binary character arrays from Netcdf Fortran to numpy object arrays"
        return np.array(
            [
                id.tobytes().decode("utf-8").strip()
                for id in self.struct_control.netcdf_file[f"{self.control_type}_id"][:]
            ],
            dtype=object,
        )

    @property
    def is_active(self) -> np.ndarray:
        return self.struct_control.netcdf_file[f"{self.control_type}_is_active"][:]

    @property
    def time(self) -> np.ndarray:
        return self.struct_control.netcdf_file[f"{self.control_type}_time"][:]

    def group_by_id(self, id: str) -> Optional[StructureControl]:
        """id is unique. Get content_type and content_pk from gridadmin. All controls are
        on lines except set_pump_capacity"""
        mask = np.isin(self.id, id)
        if np.all(~mask):
            return

        grid_id = self.grid_id[mask][0]
        action_type = self.action_type[mask][0]
        source_table, source_table_id = self.struct_control.get_source_table(
            action_type, grid_id
        )

        return StructureControl(
            id=id,
            source_table=source_table,
            source_table_id=source_table_id,
            time=self.time[mask],
            action_type=action_type,
            action_value_1=self.action_value_1[mask],
            action_value_2=self.action_value_2[mask],
            is_active=self.is_active[mask],
        )

    def group_by_action_type(self, value: str) -> List[StructureControl]:
        return self._group_by("action_type", value)

    def group_by_is_active(self, value: int) -> List[StructureControl]:
        return self._group_by("is_active", value)

    def group_by_time(self, min: float, max: float) -> List[StructureControl]:
        return self._group_by_in_between("time", min, max)

    def group_by_action_value_1(self, min: float, max: float) -> List[StructureControl]:
        return self._group_by_in_between("action_value_1", min, max)

    def group_by_action_value_2(self, min: float, max: float) -> List[StructureControl]:
        return self._group_by_in_between("action_value_2", min, max)

    def _group_by(
        self, type: str, value: Union[int, float, str]
    ) -> List[StructureControl]:
        """Group control action by a type, and sort them by unique id

        Args:
            type (str): is_active, action_type, id
            value (Union[int, float, str]): corresponding value for type

        Returns:
            List[StructureControl]: unique structures found from type, value combination
        """
        if not hasattr(self, type):
            return []

        values = getattr(self, type)
        mask = np.isin(values, value)
        ids = np.unique(self.id[mask])
        return [self.group_by_id(id) for id in ids]

    def _group_by_in_between(
        self, type: str, min: Union[int, float], max: Union[int, float]
    ) -> List[StructureControl]:
        """Find values between min and max

        Args:
            type (str): time, action_value_1, action_value_2
            value (Union[int, float, str]): corresponding value for type

        Returns:
            List[StructureControl]: unique structures found from type, value combination
        """
        if not hasattr(self, type):
            return []

        values = getattr(self, type)
        mask = (values >= min) & (values <= max)
        ids = np.unique(self.id[mask])
        return [self.group_by_id(id) for id in ids]


class GridH5WaterQualityResultAdmin(GridH5Admin):
    def __init__(
        self,
        h5_file_path: str,
        netcdf_file_path: str,
        file_modus: str = "r",
        swmr: bool = False,
    ) -> None:
        """
        :param h5_file_path: path to the hdf5 gridadmin file
        :param netcdf_file_path: path to the water quality result file
            (usually structure_control_actions_3di.nc)
        :param file_modus: modus in which to open the files
        """
        self._netcdf_file_path: str = netcdf_file_path
        super().__init__(h5_file_path, file_modus)

        if swmr:
            self.netcdf_file = H5SwmrFile(netcdf_file_path, file_modus)
        else:
            self.netcdf_file = h5py.File(netcdf_file_path, file_modus)

        self.set_timeseries_chunk_size(DEFAULT_CHUNK_TIMESERIES.stop)

        # Set substances as attributes
        substance_names = set()
        for key in self.netcdf_file.keys():
            regex_match = re.search(r"substance\d", key)
            if regex_match:
                substance_name = regex_match.group()
                if substance_name not in substance_names:
                    substance_names.add(substance_name)
                    self.__setattr__(
                        substance_name,
                        Nodes(
                            H5pyResultGroup(self.h5py_file, "nodes", self.netcdf_file),
                            **dict(
                                self._grid_kwargs,
                                **{"mixin": get_substance_result_mixin(substance_name)},
                            ),
                        ),
                    )

    def set_timeseries_chunk_size(self, new_chunk_size: int) -> None:
        """
        overwrite the default chunk size for timeseries queries.
        :param new_chunk_size <int>: new chunk size for
            timeseries queries
        :raises ValueError when the given value is less than 1
        """
        _chunk_size = int(new_chunk_size)
        if _chunk_size < 1:
            raise ValueError("Chunk size must be greater than 0")
        self._timeseries_chunk_size = slice(0, _chunk_size)
        logger.info("New chunk for timeseries size has been set to %d", new_chunk_size)
        self._grid_kwargs.update({"timeseries_chunk_size": self._timeseries_chunk_size})

    def close(self) -> None:
        super().close()
        self.netcdf_file.close()



import json
import os

import pytest
from osgeo import ogr

from threedigrid.admin.exporters.geopackage.exporter import GpkgExporter
from threedigrid.admin.lines.exporters import LinesOgrExporter

test_file_dir = os.path.join(os.getcwd(), "tests/test_files")

# the testfile is a copy of the v2_bergermeer gridadmin file
grid_admin_h5_file = os.path.join(test_file_dir, "gridadmin.h5")


@pytest.mark.parametrize("extension", [".shp", ".json"])
def test_export_by_extension(ga, tmp_path, extension):
    path = str(tmp_path / ("exporter_test_lines" + extension))
    line_2d_open_water_wgs84 = ga.lines.subset("2D_OPEN_WATER").reproject_to("4326")
    exporter = LinesOgrExporter(line_2d_open_water_wgs84)
    exporter.save(path, line_2d_open_water_wgs84.data, "4326")
    assert os.path.exists(path)
    s = ogr.Open(path)
    layer = s.GetLayer()
    assert layer.GetFeatureCount() == line_2d_open_water_wgs84.id.size


def test_nodes_gpgk_export(ga, tmp_path):
    path = str(tmp_path / ("exporter_test_lines.gpkg"))
    nodes_2d_open_water = ga.nodes.subset("2D_OPEN_WATER")
    exporter = GpkgExporter(nodes_2d_open_water)
    exporter.save(path, "nodes", ga.nodes.GPKG_DEFAULT_FIELD_MAP)
    assert os.path.exists(path)
    s = ogr.Open(path)
    layer = s.GetLayer("nodes")
    assert layer.GetFeatureCount() == nodes_2d_open_water.id.size


def test_meta_data_gpgk_export(ga, tmp_path):
    path = str(tmp_path / ("exporter_meta.gpkg"))
    exporter = GpkgExporter(ga)
    exporter.add_meta_data(path, "meta")
    assert os.path.exists(path)
    s = ogr.Open(path)
    layer = s.GetLayer("meta")
    assert layer.GetFeatureCount() == 1


def test_export_specify_fields_as_dict(ga, tmp_path):
    path = str(tmp_path / "exporter_test_lines.json")

    ga.lines.pipes.filter(content_pk=17).to_geojson(
        path, use_ogr=False, fields={"display_name": "str", "calculation_type": "str"}
    )
    with open(path) as file:
        data = json.load(file)
        assert len(data["features"][0]["properties"]) == 3
        assert "display_name" in data["features"][0]["properties"]
        assert "calculation_type" in data["features"][0]["properties"]
        assert "model_type" in data["features"][0]["properties"]


def test_export_specify_fields_as_list(ga, tmp_path):
    path = str(tmp_path / "exporter_test_lines.json")
    ga.lines.pipes.filter(content_pk=17).to_geojson(
        path, use_ogr=False, fields=["display_name", "calculation_type"]
    )
    with open(path) as file:
        data = json.load(file)
        assert len(data["features"][0]["properties"]) == 3
        assert "display_name" in data["features"][0]["properties"]
        assert "calculation_type" in data["features"][0]["properties"]
        assert "model_type" in data["features"][0]["properties"]


def test_export_specify_nested_fields(ga, tmp_path):
    path = str(tmp_path / "exporter_test_lines.json")
    ga.lines.pipes.filter(content_pk=17).to_geojson(
        path,
        use_ogr=False,
        fields=[{"first": ["display_name", {"second": ["calculation_type"]}]}],
    )
    with open(path) as file:
        data = json.load(file)
        assert len(data["features"][0]["properties"]) == 2
        assert "first" in data["features"][0]["properties"]
        assert "model_type" in data["features"][0]["properties"]
        assert "display_name" in data["features"][0]["properties"]["first"]
        assert (
            "calculation_type" in data["features"][0]["properties"]["first"]["second"]
        )


def test_export_filter(ga, tmp_path):
    path = str(tmp_path / "exporter_test_lines.json")
    ga.lines.pipes.filter(content_pk=17).to_geojson(path, use_ogr=False)
    with open(path) as file:
        data = json.load(file)
        assert len(data["features"]) == 1


def test_export_reproject(ga, tmp_path):
    path = str(tmp_path / "exporter_test_lines.json")
    ga.lines.reproject_to("4326").to_geojson(path, use_ogr=False)
    with open(path) as file:
        data = json.load(file)
        p1, _ = data["features"][1]["geometry"]["coordinates"]
        x, y = p1
        # Should be somewhere in the Netherlands
        assert 4.18 < x < 5.66
        assert 52.10 < y < 52.92


def test_export_null(ga, tmp_path):
    path = str(tmp_path / "exporter_test_null.json")
    ga.nodes.manholes.filter(content_pk=12).to_geojson(path, use_ogr=False)
    with open(path) as file:
        data = json.load(file)
        assert data["features"][0]["properties"]["drain_level"] is None


@pytest.mark.parametrize(
    "export_method,expected_filename",
    [
        ("2d_groundwater_lines", "lines_2D_groundwater"),
        ("2d_openwater_lines", "lines_2D_open_water"),
        ("2d_vertical_infiltration_lines", "lines_2D_vertical_infiltration"),
        ("levees", "levees"),
        ("breaches", "breaches"),
        ("channels", "channels"),
        ("pipes", "pipes"),
        ("weirs", "weirs"),
        ("culverts", "culverts"),
        # ("orifices", "orifices"),  no orifices in test file
        ("manholes", "manholes"),
        ("nodes", "nodes_1D_all"),
        ("pumps", "pumps"),
        ("grid", "grid_2D_open_water"),
        ("grid", "grid_2D_groundwater"),
        ("flowlines", "flowlines"),
    ],
)
def test_export_method(ga_export, export_method, expected_filename):
    getattr(ga_export, "export_" + export_method)()
    assert os.path.exists(os.path.join(ga_export._dest, expected_filename + ".json"))



"""
Models
++++++

The ``Crosssections`` models represent the cross-sections as used in the
threedicore with shape, width and height information.


"""


from threedigrid.orm.fields import ArrayField
from threedigrid.orm.models import Model


class CrossSections(Model):
    code = ArrayField(type=int)
    shape = ArrayField(type=int)
    content_pk = ArrayField(type=int)
    width_1d = ArrayField(type=float)
    offset = ArrayField(type=float)
    count = ArrayField(type=float)
    tables = ArrayField(type=float)



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.

import numpy as np

from .utils import PKMapper

DEFAULT_NULL_VALUE = -9999

FIELD_NULL_VALUES = {
    "display_name": b"",
    "code": b"",
}


def db_objects_to_numpy_array_dict(db_objects, field_names):
    numpy_array_dict = {}

    for field_name in field_names:
        numpy_array_dict[field_name] = [DEFAULT_NULL_VALUE] * len(db_objects)

    for index, db_object in enumerate(db_objects):
        for field_name in field_names:
            if "__" in field_name:
                # Allow to get attrs of child objects
                # by using field__attr
                splitted = field_name.split("__")
                db_object_attr = getattr(db_object, splitted[0])
                value = getattr(db_object_attr, splitted[1])
            else:
                value = getattr(db_object, field_name)

            if field_name == "the_geom":
                value = np.fromstring(str(value.wkb), dtype="uint8")

            # Convert unicode to str
            if isinstance(value, str):
                value = value.encode("utf8")
            if value is not None:
                numpy_array_dict[field_name][index] = value

    for key in numpy_array_dict:
        numpy_array_dict[key] = np.array(numpy_array_dict[key])

    return numpy_array_dict


def add_or_update_datasets(
    h5py_group,
    numpy_array_dict,
    field_names,
    pk,
    content_pk,
    ignore_mask=None,
    field_name_override=None,
):
    # map pk onto content_pk
    pk_mapper = PKMapper(pk, content_pk, ignore_mask)
    for field_name in field_names:
        if field_name == "pk":
            # Never store pk
            continue

        np_array = numpy_array_dict[field_name]

        null_value = FIELD_NULL_VALUES.get(field_name, DEFAULT_NULL_VALUE)

        data = pk_mapper.apply_on(np_array, null_value)

        if field_name_override:
            # By default use field_name if there is no override
            dataset_name = field_name_override.get(field_name, field_name)
        else:
            dataset_name = field_name

        if dataset_name not in list(h5py_group.keys()):
            dt = None
            if data.dtype.type is np.string_:
                if dataset_name == "code":
                    dt = "S32"
                elif dataset_name == "display_name":
                    dt = "S64"
                elif dataset_name == "shape":
                    dt = "S4"

            h5py_group.create_dataset(dataset_name, data=data, dtype=dt)
        else:
            values = h5py_group[dataset_name][:]
            mask = data != null_value
            values[mask] = data[mask]
            h5py_group[dataset_name][:] = values

    del data



from copy import deepcopy

import numpy as np

from threedigrid.admin.exporters.geopackage.exporter import GpkgExporter

# Obstacle export field definitions/mapping
OBSTACLES_FIELD_DEFINITIONS = {
    "id": "line_id",
    "dpumax": "exchange_level",
    # special field, not part of the model data but given as extra kwargs
    "cross_pix_coords__transformed": "the_geom",
}


class GeopackageExporter:
    def __init__(self, gridadmin_filename, gpkg_filename):
        """
        Usage:
            class Progress:
                def __init__(self):
                    self.percentage = 0

                def update(self, count, total):
                    if (count * 100) // total > self.percentage:
                        self.percentage = (count * 100) // total
                        print("Percentage done: %s" % self.percentage)

            progress = Progress()
            exporter = GeopackageExporter("/tmp/gridadmin.h5", "/tmp/bergermeer.gpkg")
            exporter.export(progress.update)
        """
        self.gridadmin_filename = gridadmin_filename
        self.gpkg_filename = gpkg_filename

        # Keep track of progress
        self.start = 0
        self.total_items = 0

    def export(self, progress_func=None):
        """
        progress_func should be in format:

            def progress_func(count, total):
                # count = current processed features
                # total = total feature
                pass
        """
        from threedigrid.admin.gridadmin import GridH5Admin

        with GridH5Admin(self.gridadmin_filename) as ga:
            # Replace 'coordinates' with 'node_coordinates'
            PUMPS_LINESTRING_FIELD_DEFINITIONS = deepcopy(
                ga.pumps.GPKG_DEFAULT_FIELD_MAP
            )
            del PUMPS_LINESTRING_FIELD_DEFINITIONS["coordinates"]
            PUMPS_LINESTRING_FIELD_DEFINITIONS.update({"node_coordinates": "the_geom"})

            cells = ga.cells.subset("2D_OPEN_WATER")
            pumps = ga.pumps.filter(id__gte=1)
            lines = ga.lines.filter(id__gte=1)
            nodes = ga.nodes.filter(id__gte=1)
            obstacles = ga.lines.filter(kcu=101)

            # Linestring geometry for pumps
            pumps_linestring = pumps.filter(node1_id__ne=-9999, node2_id__ne=-9999)
            obstacles_count = (
                obstacles.count if np.any(obstacles.cross_pix_coords) else 0
            )

            if progress_func:
                self.total_items = (
                    cells.count
                    + pumps.count
                    + lines.count
                    + nodes.count
                    + obstacles_count
                    + pumps_linestring.count
                )
                self.start = 0

            self.calculated_items = 0

            def internal_progress_func(item_count, item_total):
                progress_func(self.start + item_count, self.total_items)
                if item_count == item_total:
                    self.start += item_total
                    self.calculated_items += item_total

            internal_func = internal_progress_func if progress_func else None

            # pumps and pumps_linestring layers
            pumps.to_gpkg(self.gpkg_filename, progress_func=internal_func)
            pumps_linestring.to_gpkg(
                self.gpkg_filename,
                "pump_linestring",
                PUMPS_LINESTRING_FIELD_DEFINITIONS,
                progress_func=internal_func,
            )

            # Other layers
            lines.to_gpkg(
                self.gpkg_filename, layer_name="flowline", progress_func=internal_func
            )
            nodes.to_gpkg(self.gpkg_filename, progress_func=internal_func)
            cells.to_gpkg(self.gpkg_filename, progress_func=internal_func)

            if obstacles_count > 0:
                # override the value for 'cross_pixel_coords'
                cross_pix_coords__transformed = obstacles.cross_pix_coords_transformed(
                    ga.grid.transform
                )

                obstacles.to_gpkg(
                    self.gpkg_filename,
                    "obstacle",
                    OBSTACLES_FIELD_DEFINITIONS,
                    progress_func=internal_func,
                    cross_pix_coords__transformed=cross_pix_coords__transformed,
                )

            # Write meta info
            exporter = GpkgExporter(ga)
            exporter.add_meta_data(self.gpkg_filename, "meta")

            assert self.total_items == self.calculated_items



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
"""
Main entry point for threedigrid applications.
"""

import logging

import h5py
import numpy as np

from threedigrid.admin.breaches.models import Breaches
from threedigrid.admin.crosssections.models import CrossSections
from threedigrid.admin.h5py_datasource import H5pyGroup
from threedigrid.admin.levees.models import Levees
from threedigrid.admin.lines.models import Lines
from threedigrid.admin.nodes.models import Cells, EmbeddedNodes, Grid, Nodes
from threedigrid.admin.pumps.models import Pumps
from threedigrid.geo_utils import raise_import_exception, transform_bbox

try:
    import pyproj
except ImportError:
    pyproj = None

try:
    import asyncio

    import asyncio_rpc  # noqa

    asyncio_rpc_support = True
except ImportError:
    asyncio_rpc_support = False


from . import constants

logger = logging.getLogger(__name__)


class GridH5Admin:
    """
    Parses the hdf5 gridadmin file and exposes the model instances, e.g.::

        >>> ga = GridH5Admin(file_path)
        >>> ga.nodes
        >>> ga.lines
        >>> ...
    """

    def __init__(self, h5_file_path, file_modus="r", set_props=False):
        """
        :param h5_file_path: path to the gridadmin file
        :param file_modus: mode with which to open the file
            (defaults to r=READ)
        """

        self.grid_file = h5_file_path
        self.datasource_class = H5pyGroup
        self.is_rpc = False

        if type(h5_file_path) in (str, bytes) and h5_file_path.startswith("rpc://"):
            if not asyncio_rpc_support:
                raise Exception("Please reinstall this package with threedigrid[rpc]")

            from threedigrid.admin.rpc_datasource import H5RPCGroup, RPCFile

            self.h5py_file = RPCFile(h5_file_path, file_modus)
            self.datasource_class = H5RPCGroup
            self._grid_kwargs = {}
            self.is_rpc = True
            self.has_1d = True
        else:
            self.h5py_file = h5py.File(h5_file_path, file_modus)
            set_props = True

        if set_props:
            self._set_props()

        self._grid_kwargs = {"has_1d": self.has_1d}

    @property
    def grid(self):
        if self.is_rpc:
            raise Exception("RPC not available for grid")
        kwargs = self._grid_kwargs.copy()
        kwargs["n2dtot"] = self.get_from_meta("n2dtot")
        return Grid(
            self.datasource_class(self.h5py_file, "grid_coordinate_attributes"),
            **kwargs
        )

    @property
    def levees(self):
        if self.is_rpc:
            raise Exception("RPC no available for levees")

        return Levees(
            self.datasource_class(self.h5py_file, "levees"), **self._grid_kwargs
        )

    @property
    def nodes(self):
        return Nodes(
            self.datasource_class(self.h5py_file, "nodes"), **self._grid_kwargs
        )

    @property
    def nodes_embedded(self):
        return EmbeddedNodes(
            self.datasource_class(self.h5py_file, "nodes_embedded"), **self._grid_kwargs
        )

    @property
    def lines(self):
        return Lines(
            self.datasource_class(self.h5py_file, "lines"), **self._grid_kwargs
        )

    @property
    def cross_sections(self):
        return CrossSections(
            self.datasource_class(self.h5py_file, "cross_sections"), **self._grid_kwargs
        )

    @property
    def pumps(self):
        return Pumps(
            self.datasource_class(self.h5py_file, "pumps"), **self._grid_kwargs
        )

    @property
    def breaches(self):
        return Breaches(
            self.datasource_class(self.h5py_file, "breaches", gridadmin=self),
            **self._grid_kwargs
        )

    @property
    def cells(self):
        # treated as nodes
        return Cells(
            self.datasource_class(self.h5py_file, "nodes"), **self._grid_kwargs
        )

    @property
    def revision_hash(self):
        """mercurial revision hash of the model"""
        return self._to_str(self.h5py_file.attrs["revision_hash"])

    @property
    def revision_nr(self):
        """mercurial revision nr or id of the model"""
        return self._to_str(self.h5py_file.attrs["revision_nr"])

    @property
    def model_name(self):
        """name of the model the gridadmin file belongs to"""
        return self._to_str(self.h5py_file.attrs["model_name"])

    @property
    def epsg_code(self):
        return self._to_str(self.h5py_file.attrs["epsg_code"])

    @property
    def crs(self):
        if pyproj is None:
            raise_import_exception("pyproj")
        try:
            return pyproj.CRS(self._to_str(self.h5py_file.attrs["crs_wkt"]))
        except KeyError:
            # Fallback for older gridadmins without crs_wkt
            return pyproj.CRS.from_epsg(int(self.h5py_file.attrs["epsg_code"]))

    @property
    def model_slug(self):
        return self._to_str(self.h5py_file.attrs["model_slug"])

    @property
    def threedicore_version(self):
        return self._to_str(self.h5py_file.attrs["threedicore_version"])

    @property
    def threedi_version(self):
        return self._to_str(self.h5py_file.attrs["threedi_version"])

    @property
    def has_levees(self):
        if not hasattr(self, "levees"):
            return False
        return bool(self.levees.id.size)

    def get_extent_subset(self, subset_name, target_epsg_code=""):
        """

        :param subset_name: name of the node subset for which the extent
            should be calculated.
            Valid input: '1D_all', '2D_groundwater'
        :param target_epsg_code: string representation of the desired
            output epsg code
        :return: numpy array of xy-min/xy-max pairs
        """
        if self.is_rpc:
            # Proxy function via RPC
            return self.h5py_file.get_extent_subset(
                subset_name, target_espg_code=target_epsg_code
            )

        attr_name = constants.SUBSET_NAME_H5_ATTR_MAP.get(subset_name.upper())
        extent = self.h5py_file.attrs.get(attr_name, None)
        if extent is None:
            raise AttributeError(
                "The grid admin file %s does not have %s attribute"
                % (self.grid_file, attr_name)
            )
        if attr_name == constants.EXTENT_1D_KEY and not self.has_1d:
            logger.info("[*] The model has no 1D returning...")
            return
        if attr_name == constants.EXTENT_2D_KEY and not self.has_2d:
            logger.info("[*] The model has no 2D returning...")
            return

        if target_epsg_code and target_epsg_code != self.epsg_code:
            extent = transform_bbox(
                extent,
                self.epsg_code,
                target_epsg_code,
            )
        return extent

    def get_model_extent(self, target_epsg_code="", **kwargs):
        """
        get the extent of the model. Combines the extent of 1d and 2d node
        coordinates. If target_epsg_code is different from the models epsg
        code it projects the coordinates to projection of the given
        target epsg code.

        :param target_epsg_code: string representation of the desired
            output epsg code
        :param kwargs:
            extra_extent: list of xy-min/xy-max pairs that need to be
            taken into account in the model extent calculation

        :return: numpy array of xy-min/xy-max pairs

        """
        bbox = kwargs.get("extra_extent", [])
        if self.is_rpc:
            # Proxy function via RPC
            return self.h5py_file.get_model_extent(
                target_epsg_code=target_epsg_code, bbox=bbox
            )

        for k in constants.SUBSET_NAME_H5_ATTR_MAP.keys():
            sub_extent = self.get_extent_subset(
                subset_name=k, target_epsg_code=target_epsg_code
            )
            if sub_extent is None:
                continue
            bbox.append(sub_extent)

        x = np.array(bbox)[:, [0, 2]]
        y = np.array(bbox)[:, [1, 3]]

        return np.array([np.min(x), np.min(y), np.max(x), np.max(y)])

    def _set_props(self):
        if not self.is_rpc:
            for prop, value in self.h5py_file.attrs.items():
                if prop and prop.startswith("has_"):
                    try:
                        setattr(self, prop, bool(value))
                    except AttributeError:
                        logger.warning(
                            "Can not set property {}, already exists".format(prop)
                        )
                        pass
        else:
            from threedigrid.admin.rpc_datasource import _set_property

            properties = ["has_1d", "has_2d", "has_breaches", "has_pumpstations"]
            futures = [_set_property(self, prop) for prop in properties]
            return asyncio.gather(*futures)

    def get_from_meta(self, prop_name):
        if prop_name not in list(self.h5py_file["meta"].keys()):
            return None

        return self.h5py_file["meta"][prop_name][()]

    @staticmethod
    def _to_str(x):
        try:
            return x.decode("utf-8")
        except AttributeError:
            return x

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.h5py_file.close()

    def close(self):
        self.h5py_file.flush()
        self.h5py_file.close()



import re
from uuid import uuid4

from threedigrid.orm.base.datasource import DataSource

try:
    import asyncio_rpc  # noqa
except ImportError:
    raise Exception(
        "asyncio_rpc needs to be installed when using this module,"
        "install it with: pip install asyncio_rpc"
    )

from asyncio_rpc.client import RPCClient
from asyncio_rpc.commlayers.redis import RPCRedisCommLayer
from asyncio_rpc.models import RPCCall, RPCStack, RPCSubStack
from asyncio_rpc.serialization import msgpack

NAMESPACE = "GRIDRESULTADMIN"
RESULT_EXPIRE_TIME = 30


async def _set_property(ga, prop):
    value = await ga.h5py_file.attrs[prop].resolve()
    setattr(ga, prop, bool(value))


class Future:
    def __init__(self, rpc_file, rpc_stack):
        self.rpc_file = rpc_file
        self.rpc_stack = rpc_stack

    async def resolve(self):
        """
        Resolve this future
        """
        client = await self.rpc_file.client
        return await client.rpc_call(self.rpc_stack)


class FutureResult(Future):
    async def resolve(self):
        """
        Returns by default only NetCDF results
        """
        client = await self.rpc_file.client
        return await client.rpc_call(self.rpc_stack)

    async def subscribe(self, only_netcdf_results=False, max_items_per_second=None):
        """
        :param rate_limit:
            maximum number of items returned per second. Cannot
            be higher than 5

        :param only_netcdf_results:
            if True, only include results that are in the NetCDF
        """
        client = await self.rpc_file.client
        if only_netcdf_results:
            self.rpc_stack.stack.append(RPCCall("only_netcdf_results", [], {}))
        if max_items_per_second:
            self.rpc_stack.stack.append(
                RPCCall("max_items_per_second", [max_items_per_second], {})
            )
        rpc_substack = RPCSubStack(**self.rpc_stack.__dict__)
        return await client.subscribe_call(rpc_substack)


class RPCAttrs:
    def __init__(self, rpc_file):
        self.rpc_file = rpc_file

    def get(self, key, *args, **kwargs):
        # Use nodes route for now
        stack = [
            RPCCall("nodes", [], {}),
            RPCCall("_datasource", [], {}),
            RPCCall("getattr", [key], {}),
        ]

        rpc_stack = RPCStack(uuid4().hex, NAMESPACE, RESULT_EXPIRE_TIME, stack)
        return Future(self.rpc_file, rpc_stack)

    def __getitem__(self, key):
        """
        Get item from attrs
        """
        return self.get(key)


class RPCFile:
    def __init__(self, path, file_modus):
        match = re.match(r"rpc://(?P<host>\S+)/(?P<channel>\S+)", path)

        # Default value for now
        self._redis_port = 6379

        if match:
            data = match.groupdict()
            self._redis_host = data["host"]
            self.pubchannel = data["channel"]
        else:
            raise Exception("Could not parse path %s" % path)

        self.path = path
        self.file_modus = file_modus
        self._client = None

    def filepath(self):
        return self.path

    @property
    async def client(self):
        if self._client is None:
            subchannel = uuid4().hex
            comm = await RPCRedisCommLayer.create(
                subchannel,
                self.pubchannel,
                host=self._redis_host,
                port=self._redis_port,
                serialization=msgpack,
            )
            self._client = RPCClient(comm)
        return self._client

    def get_model_extent(self, target_epsg_code="", bbox=[]):
        stack = [
            RPCCall(
                "get_model_extent",
                [],
                {"target_epsg_code": target_epsg_code, "bbox": bbox},
            ),
        ]
        rpc_stack = RPCStack(uuid4().hex, NAMESPACE, RESULT_EXPIRE_TIME, stack)
        return Future(self, rpc_stack)

    def get_extent_subset(self, subset_name, target_espg_code=""):
        stack = [
            RPCCall(
                "get_extent_subset",
                [subset_name],
                {"target_espg_code": target_espg_code},
            ),
        ]
        rpc_stack = RPCStack(uuid4().hex, NAMESPACE, RESULT_EXPIRE_TIME, stack)
        return Future(self, rpc_stack)

    def __getitem__(self, key):
        if key == "meta":
            return self.meta
        else:
            raise Exception("Only [meta] is implemented")

    @property
    def meta(self):
        # Use nodes route for now
        stack = [
            RPCCall("nodes", [], {}),
            RPCCall("_datasource", [], {}),
            RPCCall("getattr", ["meta"], {}),
        ]

        rpc_stack = RPCStack(uuid4().hex, NAMESPACE, RESULT_EXPIRE_TIME, stack)
        return Future(self, rpc_stack)

    @property
    def attrs(self):
        return RPCAttrs(self)


class H5RPCGroup(DataSource):
    """
    Datasource wrapper for h5py groups,
    adding meta data to all groups.
    """

    future_class = Future

    def __init__(self, rpc_file, group_name, meta=None, required=False, gridadmin=None):
        self.rpc_file = rpc_file
        self.group_name = group_name

    def get_rpc_actions(self, model, field_name=None):
        rpc_actions = []
        if model.reproject_to_epsg:
            rpc_actions.append({"reproject_to": model.reproject_to_epsg})
        rpc_actions += [x.to_dict() for x in model.slice_filters if not x._filter_as]

        if field_name is None:
            if model.only_fields:
                rpc_actions.append({"only": model.only_fields})
            rpc_actions.append({"data": {}})
        else:
            rpc_actions.append({field_name: {}})

        return rpc_actions

    def get_asyncio_rpc_stack(self, model, rpc_actions):
        model_name = model.__class__.__name__.lower()

        stack = []

        # Override for cells here
        if model_name != "cells":
            stack.append(RPCCall(self.group_name, [], {}))

        if model_name != self.group_name:
            stack.append(RPCCall(model_name, [], {}))

        for action in rpc_actions:
            for key, value in action.items():
                if isinstance(value, list):
                    stack.append(RPCCall(key, value, {}))
                elif isinstance(value, dict):
                    stack.append(RPCCall(key, [], value))
                else:
                    stack.append(RPCCall(key, [value], {}))

        return RPCStack(uuid4().hex, NAMESPACE, RESULT_EXPIRE_TIME, stack)

    def get_filtered_field_value(
        self, model, field_name, ts_filter=None, lookup_index=None, subset_index=None
    ):
        rpc_actions = self.get_rpc_actions(model, field_name=field_name)
        rpc_stack = self.get_asyncio_rpc_stack(model, rpc_actions)

        return self.future_class(self.rpc_file, rpc_stack)

    def execute_query(self, model):
        rpc_actions = self.get_rpc_actions(model)
        rpc_stack = self.get_asyncio_rpc_stack(model, rpc_actions)

        return self.future_class(self.rpc_file, rpc_stack)

    def set(self, name, values):
        pass

    def getattr(self, name):
        stack = [
            RPCCall(self.group_name, [], {}),
            RPCCall("_datasource", [], {}),
            RPCCall("getattr", [name], {}),
        ]

        rpc_stack = RPCStack(uuid4().hex, NAMESPACE, RESULT_EXPIRE_TIME, stack)
        return self.future_class(self.rpc_file, rpc_stack)

    def keys(self):
        return []

    def has_any(self, sources):
        """
        Check if the any of the sources can be found in
        the file.
        """
        # By default pretend to have all sources,
        # an exception is thrown if the source is not available
        return True


class H5RPCResultGroup(H5RPCGroup):
    future_class = FutureResult

    def __init__(self, rpc_file, group_name, netcdf_file, meta=None, required=False):
        super().__init__(rpc_file, group_name, meta, required)
        self.netcdf_file = netcdf_file

    def get_timestamps(self):
        stack = [
            RPCCall(self.group_name, [], {}),
            RPCCall("_datasource", [], {}),
            RPCCall("get", ["time"], {}),
            RPCCall("value", [], {}),
        ]

        rpc_stack = RPCStack(uuid4().hex, NAMESPACE, RESULT_EXPIRE_TIME, stack)
        return self.future_class(self.rpc_file, rpc_stack)

    def get_rpc_actions(self, model, field_name=None):
        rpc_actions = super().get_rpc_actions(model, field_name)

        # Check if we need to sample
        if model.timeseries_sample is not None:
            rpc_actions.insert(0, {"sample": model.timeseries_sample})

        # Check if we need to inject timeseries_filter
        if model.timeseries_filter is not None:
            # insert action
            rpc_actions.insert(0, {"timeseries": model.timeseries_filter})

        return rpc_actions

    def attr(self, var_name, attr_name):
        pass



from enum import Enum
from typing import List


class StructureControlTypes(Enum):
    table_control = "table_control"
    memory_control = "memory_control"
    timed_control = "timed_control"


class StructureControl:
    def __init__(
        self,
        id: str,
        source_table: str,
        source_table_id: int,
        time: List[float],
        action_type: str,
        action_value_1: List[float],
        action_value_2: List[float],
        is_active: List[int],
    ) -> None:
        self.id = id
        self.source_table = source_table
        self.source_table_id = source_table_id
        self.time = time
        self.action_type = action_type
        self.action_value_1 = action_value_1
        self.action_value_2 = action_value_2
        self.is_active = is_active



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.

SHP_DRIVER_NAME = "ESRI Shapefile"
GEO_PACKAGE_DRIVER_NAME = "GPKG"
GEOJSON_DRIVER_NAME = "GeoJSON"

EXPORT_SHAPEFILE = "to_shape"
EXPORT_GEOPACKAGE = "to_gpkg"
EXPORT_GEOJSON = "to_geojson"

SHP_EXTENSION = ".shp"
GEO_PACKAGE_EXTENSION = ".gpkg"
GEOJSON_EXTENSION = ".json"
GEOJSON_EXTENSION2 = ".geojson"

EXTENSION_TO_DRIVER_MAP = {
    SHP_EXTENSION: SHP_DRIVER_NAME,
    GEO_PACKAGE_EXTENSION: GEO_PACKAGE_DRIVER_NAME,
    GEOJSON_EXTENSION: GEOJSON_DRIVER_NAME,
    GEOJSON_EXTENSION2: GEOJSON_DRIVER_NAME,
}

EXPORT_METHOD_TO_EXTENSION_MAP = {
    EXPORT_SHAPEFILE: SHP_EXTENSION,
    EXPORT_GEOPACKAGE: GEO_PACKAGE_EXTENSION,
    EXPORT_GEOJSON: GEOJSON_EXTENSION,
}

###############################################################################
# White list of variables and aggregation variables
AGGREGATION_OPTIONS = {
    "min",
    "max",
    "avg",
    "med",
    "cum",
    "cum_positive",
    "cum_negative",
}

###############################################################################
# result netCDF variables


# pumpstations
PUMPS_VARIABLES = ["Mesh1D_q_pump"]

# breaches
BREACH_VARIABLES = ["Mesh1D_breach_depth", "Mesh1D_breach_width"]
LEVEES_VARIABLES = []



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.

import logging
from functools import lru_cache

import numpy as np

try:
    import pyproj
except ImportError:
    pyproj = None

try:
    from osgeo import osr
except ImportError:
    osr = None

try:
    import mercantile
except ImportError:
    mercantile = None

try:
    import shapely
    from shapely.strtree import STRtree
    from shapely.wkt import loads
except ImportError:
    shapely = None

from threedigrid.numpy_utils import select_lines_by_bbox

logger = logging.getLogger(__name__)

MERCANTILE_EPSG_CODE = "4326"

BBOX_LEFT = 0
BBOX_TOP = 1
BBOX_RIGHT = 2
BBOX_BOTTOM = 3


def raise_import_exception(name):
    raise ImportError(
        "Could not import {}, you need to install threedigrid "
        "with the extra [geo], e.g "
        "pip install threedigrid[geo]==<version>".format(name)
    )


@lru_cache(10)
def get_transformer(source_epsg, target_epsg):
    if pyproj is None:
        raise_import_exception("pyproj")

    return pyproj.Transformer.from_crs(
        pyproj.CRS.from_epsg(int(source_epsg)),
        pyproj.CRS.from_epsg(int(target_epsg)),
        always_xy=True,
    )


def transform_xys(x_array, y_array, source_epsg, target_epsg):
    """
    Transform x_array, y_array from source_epsg_code to
    target_epsg code
    """
    transformer = get_transformer(source_epsg, target_epsg)
    assert isinstance(x_array, np.ndarray)
    assert isinstance(y_array, np.ndarray)

    if x_array.size == 0 and y_array.size == 0:
        return np.array([[], []])

    reprojected = transformer.transform(x_array, y_array)
    return np.array(reprojected)


def get_spatial_reference(epsg_code):
    """
    :param epsg_code: Spatial Reference System Identifier (SRID)
        as implemented by the European Petroleum Survey Group (EPSG).

    :return: an osr spatial reference instance

    :raise RuntimeError: if the import from epsg_code failed

    """
    if osr is None:
        raise_import_exception("osr")

    sr = osr.SpatialReference()
    try:
        sr.ImportFromEPSG(int(epsg_code))
        return sr
    except RuntimeError:
        logger.exception(
            "[-] Importing projection from epsg code %s failed." % epsg_code
        )
        raise


def transform_bbox(bbox, source_epsg_code, target_epsg_code, all_coords=False):
    """
    Transform bbox from source_epsg_code to target_epsg_code,
    if necessary

    :returns np.array of shape 4 which represent the two coordinates:
        left, bottom and right, top.
        When `all_coords` is set to `True`, a np.array of shape 8 is given
        which represents coords of the bbox in the following order:
        left top, right top, right bottom, left bottom
    """
    if source_epsg_code != target_epsg_code:
        # XXX: Not entirely sure whether transformations between two projected
        # coordinate systems always do retain the rectangular shape of a bbox.
        # Transformations between an unprojected system (e.g. WGS84) and a
        # projected system (e.g. RDNEW) will experience distortion: the
        # resulting shape cannot be accurately represented by top left
        # and bottom right.
        source_srs = get_spatial_reference(source_epsg_code)
        target_srs = get_spatial_reference(target_epsg_code)
        if source_srs.IsProjected() != target_srs.IsProjected():
            msg = "Transforming a bbox from %s to %s is inaccurate."
            logger.warning(msg, source_epsg_code, target_epsg_code)
        # Transform to [[left, right],[top, bottom]]
        input_x = [bbox[BBOX_LEFT], bbox[BBOX_RIGHT]]
        input_y = [bbox[BBOX_TOP], bbox[BBOX_BOTTOM]]
        if all_coords:
            input_x += [bbox[BBOX_RIGHT], bbox[BBOX_LEFT]]
            input_y += [bbox[BBOX_TOP], bbox[BBOX_BOTTOM]]
        bbox_trans = np.array(
            transform_xys(
                np.array(input_x), np.array(input_y), source_epsg_code, target_epsg_code
            )
        )

        if all_coords:
            bbox = np.array(
                [
                    bbox_trans[0][0],
                    bbox_trans[1][0],  # left_top
                    bbox_trans[0][2],
                    bbox_trans[1][2],  # right_top
                    bbox_trans[0][1],
                    bbox_trans[1][1],  # right_bottom
                    bbox_trans[0][3],
                    bbox_trans[1][3],  # left_bottom
                ]
            )
        else:
            # Transform back to [left,bottom,right,top]
            bbox = np.array(
                [
                    min(bbox_trans[0]),
                    min(bbox_trans[1]),  # left_bottom
                    max(bbox_trans[0]),
                    max(bbox_trans[1]),  # right_top
                ]
            )
    return bbox


def get_bbox_for_tile(tile_xyz=(0, 0, 0), target_epsg_code="4326"):
    """
    Get the bbox for a tile defined by x,y,z in epsg=target_epsg_code

    :param tile_xyz: tuple with tile x,y,z
    :param target_epsg_code: the epsg_code to reproject the bbox to.

    :return: The bbox of the tile
             (reprojected to target_epsg_code if necessary)
             Numpy array: [left, top, right, bottom]
             y-axis is inverted: bottom >= top

    Examples:
    >>> get_bbox_for_tile((10, 10, 15))
    array([-179.89013672,   85.04069252, -179.87915039,   85.04164217])

    >>> get_bbox_for_tile((10, 10, 15), '28992')
    array([  212572.58551456,  5468921.2451706 ,   212705.73721283,
            5469031.60724307])
    """
    if mercantile is None:
        raise_import_exception("mercantile")
    bbox = np.array(
        mercantile.bounds(
            mercantile.Tile(x=int(tile_xyz[0]), y=int(tile_xyz[1]), z=int(tile_xyz[2]))
        )
    )

    if MERCANTILE_EPSG_CODE != target_epsg_code:
        bbox = transform_bbox(bbox, MERCANTILE_EPSG_CODE, target_epsg_code)

    return bbox


def select_points_by_bbox(points, bbox):
    """
    Return a boolean mask array for the points that are in bbox

    :param points: np.array [[x1, x2, x3...], [y1, y2, y3...]]
    :param bbox: np.array [left, top, right, bottom], bottom >= top
    :return: a boolean mask array with 'True' for the
             points that are in the bbox

    Example:
    >>> select_points_by_bbox(
    ...     points=np.array([[1, 10, 1, 10], [10, 10, 50, 50]]),
    ...     bbox=np.array([10, 10, 50, 50]))
    array([False,  True, False,  True], dtype=bool)

    >>> select_points_by_bbox(
    ...     points=np.array([[-11, 10, 10, -10], [10, 10, -11, -10]]),
    ...     bbox=np.array([-10, -10, 10, 10]))
    array([False,  True, False,  True], dtype=bool)

    >>> select_points_by_bbox(
    ...     points=np.array([[-11, 10, 10, -10], [10, 10, 0, -10]]),
    ...     bbox=np.array([-10, 0, 10, 10]))
    array([False,  True,  True, False], dtype=bool)

    >>> select_points_by_bbox(
    ...     points=np.array([[-1, 0, 0, 10], [0, 0, -11, -10]]),
    ...     bbox=np.array([0, -10, 10, 0]))
    array([False,  True, False,  True], dtype=bool)

    >>> select_points_by_bbox(
    ...     points=np.array([[100, -100], [110, -110]]),
    ...     bbox=np.array([0, -10, 10, 0]))
    array([False, False], dtype=bool)
    """
    lleft = np.array([bbox[BBOX_LEFT], bbox[BBOX_TOP]])  # lower-left
    uright = np.array([bbox[BBOX_RIGHT], bbox[BBOX_BOTTOM]])  # upper-right
    xy_points = points.T
    pre_sel_low = xy_points >= lleft
    pre_sel_up = xy_points <= uright
    return np.all((pre_sel_low & pre_sel_up), axis=1)


def select_points_by_tile(tile_xyz=(0, 0, 0), target_epsg_code="4326", points=None):
    """
    Select points by a tile

    :param tile_xyz: tuple with tile x,y,z
    :param target_epsg_code: the epsg_code to reproject the bbox to
            before selecting.
    :param points: (np.array) array [[x1, x2, x3...], [y1, y2, y3...]]
    :return: A boolean mask for the points within the tile_xyz bbox.

    Example:
    >>> select_points_by_tile((10, 10, 15),
    ...     '4326',
    ...     np.array([[-179.89013670, -179.89013670],
    ...               [  85.04069252,   85.04269252]]))
    array([ True, False], dtype=bool)
    """

    assert isinstance(points, np.ndarray)

    return select_points_by_bbox(points, get_bbox_for_tile(tile_xyz, target_epsg_code))


def select_lines_by_tile(
    tile_xyz=(0, 0, 0), target_epsg_code="4326", lines=None, include_intersections=False
):
    """
    Select lines by a tile

    :param tile_xyz: tuple with tile x,y,z
    :param target_epsg_code: the epsg_code to reproject the bbox to
            before selecting.
    :param lines: (np.array) array:
                    x1_array=lines[0], y1_array=lines[1],
                    x2_array=lines[2], y2_array=lines[3]
           line[i] = (x1[i], y1[i]) - (x2[i], y2[i])

    :return: The bbox of the tile
             (reprojected to target_epsg_code if necessary)
    """

    assert isinstance(lines, np.ndarray)

    return select_lines_by_bbox(
        lines, get_bbox_for_tile(tile_xyz, target_epsg_code), include_intersections
    )


def select_geoms_by_geometry(geoms, geometry):
    """Build an STRtree from geoms and returns indices into 'geoms'
    where geometry intersects.

    :param geoms: list of geometries you want to search from
    :param geometry: intersection geometry
    :return: ndarray of indices into 'geoms'
    """
    if shapely is None:
        raise_import_exception("shapely")

    if type(geometry) in (bytes, str):
        if isinstance(geometry, bytes):
            geometry = geometry.decode("utf-8")
        # assume wkt, try to load
        geometry = loads(geometry)

    tree = STRtree(geoms)
    # STRtree checks intersection based on bbox of the geometry only:
    # https://github.com/Toblerity/Shapely/issues/558

    if shapely.__version__.startswith("1."):
        result = []

        for i, g in enumerate(geoms):
            g.index = i

        for intersected_geom in tree.query(geometry):
            if geometry.intersects(intersected_geom):
                result.append(intersected_geom.index)
        return np.array(result, dtype=int)
    else:
        return tree.query(geometry, predicate="intersects")



import logging
import os
from abc import ABCMeta, abstractmethod

from threedigrid.admin.constants import FID_FIELDS, TYPE_FUNC_MAP

try:
    from osgeo import ogr
except ImportError:
    ogr = None

try:
    from osgeo import gdal
except ImportError:
    gdal = None

from threedigrid.geo_utils import raise_import_exception
from threedigrid.orm import constants
from threedigrid.orm.base.exceptions import DriverNotSupportedError

logger = logging.getLogger(__name__)


class BaseExporterObject(metaclass=ABCMeta):
    @abstractmethod
    def save(self):
        """
        exporter objects must implement the save method
        """
        pass


class BaseOgrExporter(BaseExporterObject):
    driver = None

    def __init__(self):
        if ogr is None:
            raise_import_exception("ogr")
        if gdal is None:
            raise_import_exception("gdal")

    @staticmethod
    def set_field(feature, field_name, field_type, raw_value):
        field_name = str(field_name)
        # Try to de-numpy the dtype
        try:
            raw_value = raw_value.item()
        except AttributeError:
            pass

        value = TYPE_FUNC_MAP[field_type](raw_value)

        if value is None or value == -9999:
            feature.SetFieldNull(field_name)
        else:
            feature.SetField(field_name, value)
        if field_name in FID_FIELDS and value is not None:
            feature.SetFID(value)

    def set_driver(self, driver_name="", extension=""):
        assert any(
            (driver_name, extension)
        ), "either driver_name or extension must be given"
        if not driver_name:
            driver_name = constants.EXTENSION_TO_DRIVER_MAP[extension.lower()]
        if all(
            [
                driver_name == constants.GEO_PACKAGE_DRIVER_NAME,
                int(gdal.VersionInfo("VERSION_NUM")) < 2000000,
            ]
        ):
            logger.error("Requires GDAL >= 2.0(dev)")
            raise DriverNotSupportedError("Requires GDAL >= 2.0(dev)")
        self.driver = ogr.GetDriverByName(str(driver_name))

    @property
    def driver_name(self):
        if self.driver:
            return self.driver.GetName()

    def del_datasource(self, file_name):
        if not os.path.exists(file_name):
            return
        logger.info("[*] Replacing %s ...", file_name)
        self.driver.DeleteDataSource(str(file_name))



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.

from threedigrid.orm.fields import ArrayField, LineArrayField, PointArrayField
from threedigrid.orm.models import Model


class Pumps(Model):
    display_name = ArrayField(type=str)
    content_pk = ArrayField(type=int)
    type = ArrayField(type=int)
    node1_id = ArrayField(type=int)
    node2_id = ArrayField(type=int)
    bottom_level = ArrayField(type=float)
    start_level = ArrayField(type=float)
    lower_stop_level = ArrayField(type=float)
    capacity = ArrayField(type=float)
    coordinates = PointArrayField()
    # coordinates is the centroid of
    # node_coordinates if both set, else
    # the one that is set.
    node_coordinates = LineArrayField()
    # [ [node1_x], [node1_y], [node2_x], [node2_y]]
    # Note: -9999 if nodeX_id is -9999 (not set)
    zoom_category = ArrayField(type=int)

    GPKG_DEFAULT_FIELD_MAP = {
        "id": "id",
        "display_name": "display_name",
        "node1_id": "calculation_node_id_start",
        "node2_id": "calculation_node_id_end",
        "content_pk": "source_table_id",
        "type": "type",
        "bottom_level": "bottom_level",
        "start_level": "start_level",
        "lower_stop_level": "lower_stop_level",
        # "start_level": "upper_stop_level",
        "capacity": "capacity",
        "coordinates": "the_geom",
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
